#include    "memorymap.h"

#define	TRUE	(1)
#define FALSE	(0)
#define RUNNABLE (1)
#define WAITING (0)
#define BUFFER_SIZE 1024
#define	TASK_LIST_SIZE 30

static void MAM_init(void);
static void PLL_init(void);
static void vectored_interrupts_init(void);
static void ports_init(void);
static void serial_init(void);
static void timer_init(void);

void timer0_irq_handler(void);
void uart0_irq_handler(void);
void default_irq_handler(void);
int func_char_out(unsigned char ch);
int func_string_out(unsigned char* string);
void task_1(int what);
void task_2(int what);
void allzero(void);

typedef struct
{
	int needtorun;
	void (*task_addr)(int);
	int ID;
	int priority;
	int state;
	int argument;
} task_t;

task_t task_queue[TASK_LIST_SIZE];
task_t actual_task;
long counter=0;
int whoisrunning[TASK_LIST_SIZE];
char uart0_fifo[BUFFER_SIZE];
int uart0_fifo_state;

static void startup_init()
{
 	MAM_init();
	PLL_init();
	ports_init();
	timer_init();
	serial_init();
	vectored_interrupts_init();
}

static void MAM_init(void)
{
	REG8(SYSCON_MAMCR) 	= 0x00;  	// MAM functions disabled
	REG8(SYSCON_MAMTIM)	= 0x03;		// MAM fech time 1 processor clock
	REG8(SYSCON_MAMCR) 	= 0x02;  	// MAM functions fully enabled
}

static void PLL_init(void)
{
	unsigned char i;

        // Disabling IT
        REG32(VICIntEnClr) = 0xFFFFFFFF;

	// Set up the PLL frequency
	REG8(SYSCON_PLLCFG) = 0x25; 	// P = 2, M = 6
	REG8(SYSCON_PLLCON) = 0x01; 	// Enabling PLL but not connecting it

	REG8(SYSCON_PLLFEED) = 0xAA;	// Actualising the setting
	for(i=1;i<10;i++);			// waiting
	REG8(SYSCON_PLLFEED) = 0x55;	// Actualising the setting

	while ( !(REG16(SYSCON_PLLSTAT) & 0x0400) ) ; // waiting for PLL lock

	REG8(SYSCON_PLLCON) =  0x03; 	// Enabling PLL but not connecting it

	REG8(SYSCON_PLLFEED) = 0xAA;	// Actualising the setting
	for(i=1;i<10;i++);		// waiting
	REG8(SYSCON_PLLFEED) = 0x55;	// Actualising the setting
}

static void ports_init(void)
{
	REG32(PCB_PINSEL0) = 0x00000005;	// LED0(P2), LED1(P3), USART TX / RX
}

static void serial_init(void)
{
	REG8(UART0_LCR) = 0x83;  	// * bit character 1 stop bit no parity Enable Divisor Latches
	REG8(UART0_DLL) = 98; 		// 9600 Baud
	REG8(UART0_DLM) = 0;
	REG8(UART0_LCR) = 0x03;  	// * bit character 1 stop bit no parity Disable Divisor Latches
	REG8(UART0_FCR) = 0x01;  	// * Enabling and clearing Rx Tx FIFOs
	REG8(UART0_IER) = 0x03;		// Enabling RDA and THRE interrupts for uart0 bit 0 and bit 1
}

static void timer_init(void)
{
	// TODO
	REG8(SYSCON_VPBDIV) = 0x00;	// Divide system clock by four for the peripherals 60Mhz/4 = 15Mhz

	REG16(TIMER0_MCR) = 0x3;		// Interrupt and reset on match with MR0
	REG32(TIMER0_PR)  = 100;		// 15Mhz/100 = 0.15Mhz
	REG32(TIMER0_MR0) = 750;		// 0.15Mhz/750 = 200Hz

	REG8(TIMER0_TCR) = 0x1;			// Counter0 is enabled
}

static void vectored_interrupts_init(void)
{
	REG32(VICIntSelect) = 0x0;	// All interrupts are assigned to IRQ category
	REG32(VICIntEnable) = 0x50;	// Timer0 (bit4) uart0 (bit6) interrupt is enabled  /1010000/
	REG32(VICVectAddr0) = (unsigned long)timer0_irq_handler;	// Timer0 IRQ handler
	REG32(VICVectCntl0) = 0x24; // interrupt source is timer0 (source 4) and enable (bit 5) /1 00100/
	REG32(VICVectAddr1) = (unsigned long)uart0_irq_handler;     // UART0 IRQ handler
	REG32(VICVectCntl1) = 0x26;  // interrupt source is uart0 (source 6) and enable (bit 5) /1 00110/
	REG32(VICDefVectAddr) = (unsigned long)default_irq_handler;
}

void default_irq_handler(void)
{
	REG8(UART0_THR) = '@';
}

/*void interrupt_handler(void)
{
	void (*temp)();
	temp=(void (*))(REG32(VICVectAddr));	//get the address to run the ISR (should be wrote in assambly)
	temp();
	REG32(VICVectAddr) = 0xFFFFFFFF;
	//Need to write to VICVectAddr to clear the internal interrupt priority hardware's
	//respective interrupt flag
}*/

void uart0_irq_handler(void)
{
	switch (REG8(UART0_IIR) & 0x0E) //Mask with 1110
	{
		case (0x02) :				//In case of THRE interrupt (more often interrupt)
		{
			if (uart0_fifo[uart0_fifo_state] != 0)  //Do we have something to write out?
				{
					REG8(UART0_THR) = uart0_fifo[uart0_fifo_state];
					uart0_fifo[uart0_fifo_state] = 0;
					if (uart0_fifo_state==(BUFFER_SIZE-1)) {uart0_fifo_state=0;}
					else {uart0_fifo_state++;}
				}
			else
				{
					REG8(UART0_IER) = 0x01; //If we have nothing to write out
				}
			break;
		}
		case (0x04) :				//In case od RDA interrupt (less often interrupt)
		{
			REG8(UART0_THR) = REG8(UART0_RBR);
			int i;
			long sum=0;
			for(i=0;i<BUFFER_SIZE;i++){uart0_fifo[i]=0;}
			for(i=0;i<TASK_LIST_SIZE;i++){sum+=whoisrunning[i];}
			func_string_out("--The sum counting=");
			func_char_out((sum / 100000)%10+'0');
			func_char_out((sum / 10000)%10+'0');
			func_char_out((sum / 1000)%10+'0');
			func_char_out((sum / 100)%10+'0');
			func_char_out((sum / 10)%10+'0');
			func_char_out((sum % 10)+'0');
			for(i=0;i<TASK_LIST_SIZE;i++)
			{
				if (whoisrunning[i])
					{
						func_char_out(':');
						func_char_out((whoisrunning[i] / 100000)%10+'0');
						func_char_out((whoisrunning[i] / 10000)%10+'0');
						func_char_out((whoisrunning[i] / 1000)%10+'0');
						func_char_out((whoisrunning[i] / 100)%10+'0');
						func_char_out((whoisrunning[i] / 10)%10+'0');
						func_char_out((whoisrunning[i] % 10)+'0');
					}
			}
			func_char_out(':');
			break;
		}
	}
}

void timer0_irq_handler(void)
{
//	REG8(UART0_THR) = 't';
	counter++;
	whoisrunning[actual_task.ID]=whoisrunning[actual_task.ID]+1;
	if (counter==5000)
	{
		func_string_out("..counter to zero..");
		counter=0;
		allzero();
	}
	REG8(TIMER0_IR) = 0xFF;		// Clear the interrupt flag

}

void func_fiq_handler(void)
{
}

int func_char_out(unsigned char ch)
{
	int i,j=FALSE,k;

	for(i=0;((j==FALSE)&&(i<BUFFER_SIZE));i++)

	{
		k=(uart0_fifo_state+i)%BUFFER_SIZE;
		if (uart0_fifo[k]==0)
		{
			uart0_fifo[k]=ch;
			j=TRUE;
			if (!(REG8(UART0_IER) & 0x02)){REG8(UART0_IER) = 0x03;}
		}
	}

	return j;

}

int func_string_out(unsigned char* string)
{
  int i;
  for(i=0;string[i]!=0;i++) {
   if (!func_char_out(string[i])){return FALSE;}
   }
  return TRUE;
}

void allzero(void)
{
	int i;
	for(i=0;i<TASK_LIST_SIZE;i++){whoisrunning[i]=0;}
}

int put_task(void(t_addr)(int),int t_ID,int pri,int st, int arg)
{
	int i;
	 for(i=0;i<TASK_LIST_SIZE;i++)
	 {
		 if (!task_queue[i].needtorun)
		 {
			 task_queue[i].needtorun=TRUE;
			 task_queue[i].task_addr=t_addr;
			 task_queue[i].ID=t_ID;
			 task_queue[i].priority=pri;
			 task_queue[i].state=st;
			 task_queue[i].argument=arg;
			 return TRUE;
		 }
	 }
	 return FALSE;
}

void task_1(int what)
{
	switch (what)
	{
		case 0 : func_char_out('1');break;
	 	case 1 :
	 	{
			 while(!func_char_out('1')){};
			 put_task(task_2,2,5,RUNNABLE,1);
			 break;
	 	}
	}

}

void task_2(int what)
{
	switch (what)
	{
		case 0 : func_char_out('2');break;
		case 1 :
		{
					 while(!func_char_out('2')){};
					 put_task(task_1,1,5,RUNNABLE,1);
					 put_task(task_2,2,5,RUNNABLE,0);
					 break;
	 	}
	}

}


void user_task(int i, int j)
{
	int k;
	for(k=0;k<i;k++){put_task(task_1,1,5,RUNNABLE,1);}
	for(k=0;k<j;k++){put_task(task_1,1,5,RUNNABLE,0);}
}

int main (void)
{
	int i;
	startup_init();
	allzero();
	REG8(UART0_THR) = 'S';
	for(i=0;i<BUFFER_SIZE;i++){uart0_fifo[i]=0;}
	uart0_fifo_state=0;
	actual_task.ID=0;
	func_string_out("What?");
	user_task(3,2);
	while(1)
	{
		for(i=0;i<TASK_LIST_SIZE;i++)
			{
				if (task_queue[i].needtorun)
				{
					actual_task=task_queue[i];
					task_queue[i].task_addr(task_queue[i].argument);
					task_queue[i].needtorun=FALSE;
					actual_task.ID=0;
				}
			}
	}
 }

