/*******************************************************************************

********************************************************************************/
.extern main

.equ VICVectAddr, 0xFFFFF030

.text
.func       _int_vectors

// *****************************************************************************
//          Exception vector tablecommon to all ARM-based systems              
// ***************************************************************************** 
// Common to all ARM-based systems. See ARM Architecture Reference Manual,     
// Programmer's Model section for details. Table entries just jump to handlers 
// (using full 32-bit addressing).                                             
// *****************************************************************************
_int_vectors:   
    
    ldr     pc, do_reset_addr
    ldr     pc, do_undefined_instruction_addr
    ldr     pc, do_software_interrupt_addr
    ldr     pc, do_prefetch_abort_addr
    ldr     pc, do_data_abort_addr
    .long       0xB8A06F60                      /* ARM-reserved vector */
    ldr     pc, os_cpu_irq_isr_addr
    ldr     pc, os_cpu_fiq_isr_addr

do_reset_addr:                  .long   do_reset
do_undefined_instruction_addr:  .long   do_undefined_instruction
do_software_interrupt_addr:     .long   do_software_interrupt
do_prefetch_abort_addr:         .long   do_prefetch_abort
do_data_abort_addr:             .long   do_data_abort
os_cpu_irq_isr_addr:            .long   os_cpu_irq_isr
os_cpu_fiq_isr_addr:            .long   os_cpu_fiq_isr

// *****************************************************************************
//                         Yet to be implemented exceptions                    
// *****************************************************************************
// Just fall through to reset exception handler for now
// *****************************************************************************
do_undefined_instruction:
	mov     r0, #0xE0000000         //Write "U" to USART0 means undefined instruction
	add 	r0, r0, #0xC000
	mov     r1, #'U'
	strb	r1, [r0]
do_software_interrupt:
	mov     r0, #0xE0000000         //Write "I" to USART0 means software interrupt
	add 	r0, r0, #0xC000
	mov     r1, #'I'
	strb	r1, [r0]
do_prefetch_abort:
	mov     r0, #0xE0000000         //Write "A" to USART0 means prefetch abort
	add 	r0, r0, #0xC000
	mov     r1, #'A'
	strb	r1, [r0]
do_data_abort:
	mov     r0, #0xE0000000         //Write "D" to USART0 means data abort
	add 	r0, r0, #0xC000
	mov     r1, #'D'
	strb	r1, [r0]


// *****************************************************************************
//                            System reset handler
// *****************************************************************************
do_reset:

    // Set stack pointers for all modes used (supervisor, IRQ and FIQ)
    	msr     cpsr_c, #0xd3       // Ensure we're in supervisor mode
    	ldr     sp, =__stack_svc    // Set supervisor mode stack pointer

    	msr     cpsr_c, #0xd2       // Enter IRQ mode, with interrupts disabled
    	ldr     sp, =__stack_irq    // Set IRQ mode stack pointer

    	msr     cpsr_c, #0xd1       // Enter FIQ mode, with interrupts disabled
    	ldr     sp, =__stack_fiq    // Set FIQ mode stack pointer

    // Clear uninitialized data section (bss)
    	ldr     r4, =__start_bss    // First address
    	ldr     r5, =__end_bss      // Last  address    
    	mov     r6, #0x0                                

loop_zero:
    	str     r6, [r4]                                
    	add     r4, r4, #0x4                            
    	cmp     r4, r5                                  
    	blt     loop_zero

    // Copy initialized data sections into RAM
    	ldr     r4, =_fdata      // Destination address
    	ldr     r5, =_edata      // Last  address
    	ldr     r6, =_etext      // Source address
    	cmp     r4, r5                                  
    	beq     skip_initialize

loop_initialize:
    	ldr     r3, [r6]
    	str     r3, [r4]                                
    	add     r4, r4, #0x4
    	add     r6, r6, #0x4
    	cmp     r4, r5                                  
    	blt     loop_initialize
    
skip_initialize:    

#if !defined (ARM_RDP_MONITOR) && !defined (ARM_RDI_MONITOR)
	mov	r0, #0              //  No arguments
	mov	r1, #0              //  No argv either
#else
    // Need to set up standard file handles
    //	bl      initialize_monitor_handles
#endif    
    
    // Enter supervisor mode (interrupts are still disabled) and branch to start of 'C' code
	msr	cpsr_c, #0x13       /* I=0 F=0 T=0 MODE=supervisor */
	bl	main
    
.size  _int_vectors,.-_int_vectors;
.endfunc

os_cpu_irq_isr:
	stmfd	sp!, {r0-r12, lr}       //Save registry from r0-r12 and link register to stack

//	mov     r0, #0xE0000000         //Write a caracter to USART0
//	add 	r0, r0, #0xC000
//	mov     r1, #'a'
//	strb	r1, [r0]

//	bl	interrupt_handler;

	mvn	r3, #0xFC0              //get the value from VICVectAddr
	sub	r3, r3, #0xF
	ldr	r3, [r3]		//Get te walue what it holds
	
	mov	lr, pc
	mov	pc, r3			//Run what is on that adress
	
	mvn	r3, #0xFC0              //get the value from VICVectAddr
	sub	r3, r3, #0xF
	mvn	r1, #0x0		//Put all ones
	str	r1, [r3]		//to VICVectAddr (as the manula sais)
	
//	mov     r0, #0xE0000000         //Write a caracter to USART0
//	add 	r0, r0, #0xC000
//	mov     r1, #'d'
//	strb	r1, [r0]

	ldmfd	sp!, {r0-r12, lr}      //Load registry from stack
	subs	pc, lr, #4             // Return from IRQ

os_cpu_fiq_isr:
	stmfd	sp!, {r0-r7, lr}        //Save registry from r0-r7 (other registries have shadows) and link register to stack
	ldmfd	sp!, {r0-r7, lr}        //Load registry from stack
	subs	pc, lr, #4              // Return from IRQ

.align(4)
